(*
 * This is the base part of the build system.
 * Each file in the system is represented as a node.
 * Node may be virtual: the node may exist before the file
 * does.  For each file, we maintain stat and MD5 information
 * (if they exist).
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2003 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)
open Lm_printf

open Lm_location
open Lm_string_util
open Lm_filename_util

open Fmarshal

open Omake_ir
open Omake_state
open Omake_marshal
open Omake_node_type
open Omake_node_sig

(*
 * Sets and tables.
 *)
module DirCompare =
struct
   type t = dir

   let rec compare dir1 dir2 =
      match dir1, dir2 with
         DirRoot (hash1, root1), DirRoot (hash2, root2) ->
            if hash1 = hash2 then
               Pervasives.compare root1 root2
            else if hash1 < hash2 then
               -1
            else
               1
       | DirSub (hash1, key1, _, parent1), DirSub (hash2, key2, _, parent2) ->
            if hash1 = hash2 then
               let cmp = Pervasives.compare key1 key2 in
                  if cmp = 0 then
                     compare parent1 parent2
                  else
                     cmp
            else if hash1 < hash2 then
               -1
            else
               1
       | DirRoot _, DirSub _ ->
            -1
       | DirSub _, DirRoot _ ->
            1

   let rec compare_alpha dir1 dir2 =
      match dir1, dir2 with
         DirRoot (hash1, root1), DirRoot (hash2, root2) ->
            if hash1 = hash2 then
               Pervasives.compare root1 root2
            else if hash1 < hash2 then
               -1
            else
               1
       | DirSub (_, _, name1, parent1), DirSub (_, _, name2, parent2) ->
            let cmp = Pervasives.compare name1 name2 in
               if cmp = 0 then
                  compare parent1 parent2
               else
                  cmp
       | DirRoot _, DirSub _ ->
            -1
       | DirSub _, DirRoot _ ->
            1
end

module DirSet = Lm_set.LmMake (DirCompare)
module DirTable = Lm_map.LmMake (DirCompare)

module NodeCompare =
struct
   type t = node

   let compare_flags flag1 flag2 =
      match flag1, flag2 with
         NodeIsOptional, NodeIsOptional
       | NodeIsExisting, NodeIsExisting
       | NodeIsSquashed, NodeIsSquashed ->
            0
       | NodeIsOptional, NodeIsExisting
       | NodeIsOptional, NodeIsSquashed
       | NodeIsExisting, NodeIsSquashed ->
            -1
       | NodeIsExisting, NodeIsOptional
       | NodeIsSquashed, NodeIsOptional
       | NodeIsSquashed, NodeIsExisting ->
            1

   let rec compare node1 node2 =
      if node1 == node2 then
         0
      else
         match node1, node2 with
            NodeFile (hash1, dir1, key1, _), NodeFile (hash2, dir2, key2, _)
          | NodePhonyDir (hash1, dir1, key1, _), NodePhonyDir (hash2, dir2, key2, _) ->
               if hash1 = hash2 then
                  let cmp = DirCompare.compare dir1 dir2 in
                     if cmp = 0 then
                        Pervasives.compare key1 key2
                     else
                        cmp
               else if hash1 < hash2 then
                  -1
               else
                  1
          | NodePhonyGlobal (hash1, name1), NodePhonyGlobal (hash2, name2) ->
               if hash1 = hash2 then
                  Pervasives.compare name1 name2
               else if hash1 < hash2 then
                  -1
               else
                  1
          | NodePhonyFile (hash1, dir1, key1, _, name1), NodePhonyFile (hash2, dir2, key2, _, name2) ->
               if hash1 = hash2 then
                  let cmp = DirCompare.compare dir1 dir2 in
                      if cmp = 0 then
                          let cmp = Pervasives.compare key1 key2 in
                              if cmp = 0 then
                                  Pervasives.compare name1 name2
                              else
                                 cmp
                      else
                         cmp
               else if hash1 < hash2 then
                  -1
               else
                  1
          | NodeFlagged (flag1, node1), NodeFlagged (flag2, node2) ->
               if flag1 = flag2 then
                  compare node1 node2
               else
                  compare_flags flag1 flag2
          | NodeFile _,        NodePhonyGlobal _
          | NodeFile _,        NodePhonyDir _
          | NodeFile _,        NodePhonyFile _
          | NodeFile _,        NodeFlagged _
          | NodePhonyGlobal _, NodePhonyDir _
          | NodePhonyGlobal _, NodePhonyFile _
          | NodePhonyGlobal _, NodeFlagged _
          | NodePhonyDir _,    NodePhonyFile _
          | NodePhonyDir _,    NodeFlagged _
          | NodePhonyFile _,   NodeFlagged _ ->
               -1
          | NodeFlagged _,      NodeFile _
          | NodePhonyGlobal _,  NodeFile _
          | NodePhonyDir _,     NodeFile _
          | NodePhonyFile _,    NodeFile _
          | NodeFlagged _,      NodePhonyGlobal _
          | NodePhonyDir _,     NodePhonyGlobal _
          | NodePhonyFile _,    NodePhonyGlobal _
          | NodeFlagged _,      NodePhonyDir _
          | NodePhonyFile _,    NodePhonyDir _
          | NodeFlagged _,      NodePhonyFile _ ->
               1

   let rec compare_alpha node1 node2 =
      match node1, node2 with
         NodeFile (_, dir1, key1, _), NodeFile (_, dir2, key2, _)
       | NodePhonyDir (_, dir1, key1, _), NodePhonyDir (_, dir2, key2, _) ->
            let cmp = DirCompare.compare_alpha dir1 dir2 in
               if cmp = 0 then
                  Pervasives.compare key1 key2
               else
                  cmp
       | NodePhonyGlobal (_, name1), NodePhonyGlobal (_, name2) ->
            Pervasives.compare name1 name2
       | NodePhonyFile (_, dir1, key1, _, name1), NodePhonyFile (_, dir2, key2, _, name2) ->
            let cmp = DirCompare.compare_alpha dir1 dir2 in
               if cmp = 0 then
                  let cmp = Pervasives.compare key1 key2 in
                     if cmp = 0 then
                        Pervasives.compare name1 name2
                     else
                        cmp
               else
                  cmp
       | NodeFlagged (flag1, node1), NodeFlagged (flag2, node2) ->
            if flag1 = flag2 then
               compare_alpha node1 node2
            else
               compare_flags flag1 flag2
       | NodeFile _,        NodePhonyGlobal _
       | NodeFile _,        NodePhonyDir _
       | NodeFile _,        NodePhonyFile _
       | NodeFile _,        NodeFlagged _
       | NodePhonyGlobal _, NodePhonyDir _
       | NodePhonyGlobal _, NodePhonyFile _
       | NodePhonyGlobal _, NodeFlagged _
       | NodePhonyDir _,    NodePhonyFile _
       | NodePhonyDir _,    NodeFlagged _
       | NodePhonyFile _,   NodeFlagged _ ->
            -1
       | NodeFlagged _,      NodeFile _
       | NodePhonyGlobal _,  NodeFile _
       | NodePhonyDir _,     NodeFile _
       | NodePhonyFile _,    NodeFile _
       | NodeFlagged _,      NodePhonyGlobal _
       | NodePhonyDir _,     NodePhonyGlobal _
       | NodePhonyFile _,    NodePhonyGlobal _
       | NodeFlagged _,      NodePhonyDir _
       | NodePhonyFile _,    NodePhonyDir _
       | NodeFlagged _,      NodePhonyFile _ ->
            1
end

module NodeSet = Lm_set.LmMake (NodeCompare)
module NodeTable = Lm_map.LmMake (NodeCompare)

(*
 * Hashing.
 *)
module type HashSig =
sig
   val hash_root : root -> int
   val hash_dir : dir -> string -> int
   val hash_file : dir -> string -> int
   val hash_global_phony : string -> int
   val hash_file_phony : dir -> string -> string -> int
end

module Hash : HashSig =
struct
   let hash_root root =
      Hashtbl.hash root

   let hash_of_dir dir =
      match dir with
         DirRoot (hash, _)
       | DirSub (hash, _, _, _) ->
            hash

   let hash_dir parent key =
      let hash = hash_of_dir parent in
         Hashtbl.hash (hash, key)

   let hash_file dir key =
      let hash = hash_of_dir dir in
         Hashtbl.hash (hash, key)

   let hash_global_phony name =
      Hashtbl.hash name

   let hash_file_phony dir key name =
      let hash = hash_of_dir dir in
         Hashtbl.hash (hash, key, name)
end

(*
 * Get a pathname from a directory.
 * The name must be reversed.
 *)
let path_of_dir dir =
   let rec path_of_dir keypath path dir =
      match dir with
         DirRoot (_, root) ->
            root, keypath, path
       | DirSub (_, key, name, parent) ->
            path_of_dir (key :: keypath) (name :: path) parent
   in
      path_of_dir [] [] dir

(*
 * Build a list of the directories, in reverse order.
 *)
let dir_list_of_dir dir =
   let rec dir_list_of_dir path dir =
      match dir with
         DirRoot _ ->
            dir :: path
       | DirSub (_, _, _, parent) ->
            dir_list_of_dir (dir :: path) parent
   in
      dir_list_of_dir [] dir

(*
 * Produce a path (a string list) from the dir list.
 *)
let rec path_of_dir_list dirs =
   match dirs with
      DirSub (_, _, name, _) :: dirs ->
         name :: path_of_dir_list dirs
    | DirRoot _ :: dirs ->
         raise (Invalid_argument "path_of_dir_list")
    | [] ->
         []

(*
 * Make a directory node from the pathname.
 *)
let make_dir root path =
   let rec make parent path =
      match path with
         [] ->
            parent
       | name :: path ->
            let key = Lm_filename_util.normalize_string name in
            let hash = Hash.hash_dir parent key in
            let dir = DirSub (hash, key, name, parent) in
               make dir path
   in
   let hash = Hash.hash_root root in
   let root = DirRoot (hash, root) in
      make root path

(*
 * Make a file from the directory and the path.
 *)
let make_file dir name =
   let key = Lm_filename_util.normalize_string name in
   let hash = Hash.hash_file dir key in
      NodeFile (hash, dir, key, name)

(*
 * Get the current absolute name of the working directory.
 *)
let getcwd () =
   let cwd = Unix.getcwd () in
      match Lm_filename_util.filename_path cwd with
         AbsolutePath (root, dir) ->
            make_dir root dir
       | RelativePath dir ->
            raise (Invalid_argument "Unix.getcwd returned a relative path")

(*
 * A null root directory for globals.
 *)
let null_root =
   make_dir null_root []

(*
 * Split the directory name into a path.
 *)
let new_path dir path =
   let root, _, dir = path_of_dir dir in
   let path = Lm_filename_util.filename_path path in
   let root, stack, path =
      match path with
         AbsolutePath (root, path) ->
            (* This is an absolute path, so ignore the directory *)
            root, [], path
       | RelativePath path ->
            (* This is relative to the directory *)
            root, List.rev dir, path
   in
   let rec simplify stack path =
      match path with
         "" :: path
       | "." :: path ->
            simplify stack path
       | ".." :: path ->
            let stack =
               match stack with
                  _ :: stack ->
                     stack
                | [] ->
                     stack
            in
               simplify stack path
       | name :: path ->
            simplify (name :: stack) path
       | [] ->
            root, stack
   in
      simplify stack path

let new_dir dir path =
   let root, stack = new_path dir path in
      make_dir root (List.rev stack)

let new_file dir path =
   let root, stack = new_path dir path in
   let dir, file =
      match stack with
         [] ->
            make_dir root [], "."
       | name :: stack ->
            make_dir root (List.rev stack), name
   in
      make_file dir file

(*
 * Check if .. works in a directory.
 *)
let dotdot_table = ref DirTable.empty

(*
 * Force dotdot to fail.
 *)
let make_dotdot_fail dir =
   dotdot_table := DirTable.add !dotdot_table dir true

let abs_dir_name dir =
   let buf = Buffer.create 17 in
   let rec name dir =
      match dir with
         DirRoot (_, root) ->
            Buffer.add_string buf (Lm_filename_util.string_of_root root)
       | DirSub (_, key, _, (DirRoot _ as parent)) ->
            name parent;
            Buffer.add_string buf key
       | DirSub (_, key, _, parent) ->
            name parent;
            Buffer.add_char buf Lm_filename_util.separator_char;
            Buffer.add_string buf key
   in
   let () = name dir in
      Buffer.contents buf

let dotdot_fails dir =
   let table = !dotdot_table in
      try DirTable.find table dir with
         Not_found ->
            let name = abs_dir_name dir in
            let islink =
               try (Unix.lstat name).Unix.st_kind = Unix.S_LNK with
                  Unix.Unix_error _ ->
                     false
            in
            let table = DirTable.add table dir islink in
               dotdot_table := table;
               islink

(*
 * Produce string filename for the path,
 * relative to a particular directory.
 *
 * Algorithm:
 *    1. Compute the common prefix between the directory and file
 *    2. Add as many ".." as there are remaining names in the directory
 *       and concatenate the rest of the path.
 *
 * Example:
 *       dir = /a/b/c/d
 *       path = /a/b/e/f/g
 *       result = ../../e/f/g
 *)

(*
 * Create a string from the list of strings.
 *)
let rec flatten_generic (add_string : 'a -> string -> 'a) (contents : 'a -> string) (buf : 'a) (path : string list) =
   match path with
      [path] ->
         let buf = add_string buf path in
            contents buf
    | [] ->
         contents buf
    | name :: path ->
         let buf = add_string buf name in
         let buf = add_string buf Lm_filename_util.separator_string in
            flatten_generic add_string contents buf path

(*
 * Add .. to get out of the directory and down into the path.
 *)
let updirs_generic add_string contents buf dirs1 dirs2 =
   (* Abort if any of the dotdots fail *)
   if List.exists dotdot_fails dirs1 then
      None
   else
      (* Prepend the .. sequence *)
      let rec updirs dirs path =
         match dirs with
            _ :: dirs ->
               updirs dirs (".." :: path)
          | [] ->
               path
      in
      let path = path_of_dir_list dirs2 in
      let path = updirs dirs1 path in
         Some (flatten_generic add_string contents buf path)

(*
 * Compute the path of dir2 relative to dir1.
 *)
let rec relocate_generic add_string contents buf dirs1 dirs2 =
   match dirs1, dirs2 with
      DirSub (_, key1, _, _) :: dirs1', DirSub (_, key2, _, _) :: dirs2'  ->
         if key1 = key2 then
            relocate_generic add_string contents buf dirs1' dirs2'
         else
            updirs_generic add_string contents buf dirs1 dirs2
    | [], _ ->
         Some (flatten_generic add_string contents buf (path_of_dir_list dirs2))
    | _, [] ->
         updirs_generic add_string contents buf dirs1 dirs2
    | DirRoot _ :: _, _
    | _, DirRoot _ :: _ ->
         raise (Invalid_argument "relocate_generic")

(*
 * If the files differ in the root directory, just use the absolute path.
 *)
let relocate_generic add_string contents buf dir1 dir2 =
   let dirs1 = dir_list_of_dir dir1 in
   let dirs2 = dir_list_of_dir dir2 in
      match dirs1, dirs2 with
         DirRoot (_, root1) :: dirs1, DirRoot (_, root2) :: dirs2 ->
            begin
               let s =
                  if dirs1 = [] || root1 <> root2 then
                     None
                  else
                     relocate_generic add_string contents buf dirs1 dirs2
               in
                  match s with
                     Some s ->
                        s
                   | None ->
                        let buf = add_string buf (Lm_filename_util.string_of_root root2) in
                        let path2 = path_of_dir_list dirs2 in
                           flatten_generic add_string contents buf path2
            end
       | _ ->
            raise (Invalid_argument "relocate_generic")

(*
 * Directory versions.
 *)
let dir_buffer = Buffer.create 17

let dir_add_string buf s =
   Buffer.add_string buf s;
   buf

let dir_contents buf =
   let s = Buffer.contents buf in
      Buffer.clear buf;
      s

let flatten_dir dir =
   flatten_generic dir_add_string dir_contents dir_buffer dir

let relocate_dir dir1 dir2 =
   relocate_generic dir_add_string dir_contents dir_buffer dir1 dir2

(*
 * File version.
 *)
let file_contents name buf =
   let buf =
      if Buffer.length buf = 0 then
         buf
      else
         dir_add_string buf Lm_filename_util.separator_string
   in
   let buf = dir_add_string buf name in
      dir_contents buf

let flatten_file dir name =
   let buf = dir_add_string dir_buffer Lm_filename_util.separator_string in
      flatten_generic dir_add_string (file_contents name) buf dir

let relocate_file dir1 dir2 name =
   relocate_generic dir_add_string (file_contents name) dir_buffer dir1 dir2

(************************************************************************
 * Modules.
 *)
module Dir =
struct
   type t = dir

   (*
    * We assume the cwd does not change
    * once we first get it.
    *)
   let cwd_ref = ref (getcwd ())

   (*
    * Default is current working directory.
    *)
   let cwd () = !cwd_ref

   let reset_cwd () =
      let cwd = getcwd () in
         cwd_ref := getcwd ();
         make_dotdot_fail cwd

   (*
    * Building a new path.
    *)
   let chdir dir1 dir2 =
      new_dir dir1 (Lm_filename_util.unescape_string dir2)

   (*
    * Name, relative to the cwd.
    *)
   let name dir1 dir2 =
      let s = relocate_dir dir1 dir2 in
         if s = "" then
            "."
         else
            s

   (*
    * Name relative to the root.
    *)
   let fullname dir =
      name !cwd_ref dir

   (*
    * Absolute name.
    *)
   let root = make_dir Lm_filename_util.null_root []

   let absname dir =
      name root dir

   (*
    * Library directory is relative to the root.
    *)
   let lib =
      match Lm_filename_util.filename_path Omake_state.lib_dir with
         AbsolutePath (root, dir) ->
            make_dir root dir
       | RelativePath _ ->
            raise (Invalid_argument ("Omake_node.lib_dir specified as relative path: " ^ Omake_state.lib_dir))

   (*
    * home directory is also relative to the root.
    *)
   let home =
      match Lm_filename_util.filename_path home_dir with
         AbsolutePath (root, dir) ->
            make_dir root dir
       | RelativePath _ ->
            raise (Invalid_argument ("Omake_node.home_dir specified as relative path: " ^ home_dir))

   let () = make_dotdot_fail home

   (*
    * Equality.
    *)
   let equal dir1 dir2 =
      DirCompare.compare dir1 dir2 = 0

   (*
    * Marshaling.
    *)
   let marshal_root root =
      match root with
         NullRoot ->
            Magic NullRootMagic
       | DriveRoot c ->
            List [Magic DriveRootMagic; Char c]

   let unmarshal_root l =
      match l with
         Magic NullRootMagic ->
            NullRoot
       | List [Magic DriveRootMagic; Char c] ->
            DriveRoot c
       | _ ->
            raise MarshalError

   let rec marshal dir =
      match dir with
         DirRoot (hash, root) ->
            List [Magic DirRootMagic; Int hash; marshal_root root]
       | DirSub (hash, key, name, parent) ->
            List [Magic DirSubMagic; Int hash; String key; String name; marshal parent]

   let rec unmarshal l =
      match l with
         List [Magic DirRootMagic; Int hash; root] ->
            DirRoot (hash, unmarshal_root root)
       | List [Magic DirSubMagic; Int hash; String key; String name; parent] ->
            DirSub (hash, key, name, unmarshal parent)
       | _ ->
            raise MarshalError
end

(*
 * Nodes.
 *)
module Node =
struct
   type t = node
   type dir = Dir.t
   type db = node NodeTable.t

   (*
    * Create a database that remembers the names
    * so that we get some sharing.  The sharing
    * does not have to be preserved for correctness.
    *)
   type base =
      { mutable base_nodes : node NodeTable.t }

   let base =
      { base_nodes = NodeTable.empty }

   (*
    * Marshal the node database.
    *)
   let marshal_base () = base.base_nodes

   (*
    * Re-intern, after marshaling.
    *)
   let unmarshal_base db =
      let nodes = base.base_nodes in
      let rec intern_node nodes node =
         match node with
            NodeFile _
          | NodePhonyGlobal _
          | NodePhonyDir _ ->
               NodeTable.filter_add nodes node (function
                  Some old -> old
                | None -> node)
          | NodePhonyFile _ ->
               nodes
          | NodeFlagged (_, node) ->
               intern_node nodes node
      in
      let nodes =
         NodeTable.fold (fun nodes node _ -> intern_node nodes node) nodes db
      in
         base.base_nodes <- nodes

   (*
    * Create a phony name.
    *)
   let phony_global name =
      let hash = Hash.hash_global_phony name in
      let node = NodePhonyGlobal (hash, name) in
      let nodes = base.base_nodes in
         try NodeTable.find nodes node with
            Not_found ->
               base.base_nodes <- NodeTable.add nodes node node;
               node

   (*
    * Create a phony from a dir.
    *)
   let phony_dir dir name =
      let key = Lm_filename_util.normalize_string name in
      let hash = Hash.hash_file dir key in
      let node = NodePhonyDir (hash, dir, key, name) in
      let nodes = base.base_nodes in
         try NodeTable.find nodes node with
            Not_found ->
               base.base_nodes <- NodeTable.add nodes node node;
               node

   (*
    * Create a phony with a new directory.
    *)
   let phony_chdir node dir =
      match node with
         NodePhonyDir (_, _, _, name) ->
            phony_dir dir name
       | _ ->
            node

   (*
    * Create a new phony node from a previous node.
    * These are not interned.
    *)
   let rec phony_node node name =
      match node with
         NodeFile (_, dir, key1, name1) ->
            let hash = Hash.hash_file_phony dir key1 name in
               NodePhonyFile (hash, dir, key1, name1, name)
       | NodePhonyGlobal (_, name1) ->
            let key1 = Lm_filename_util.normalize_string name1 in
            let hash = Hash.hash_file_phony null_root key1 name in
               NodePhonyFile (hash, null_root, key1, name1, name)
       | NodePhonyDir (_, dir, key1, name1)
       | NodePhonyFile (_, dir, key1, name1, _) ->
            let hash = Hash.hash_file_phony dir key1 name in
               NodePhonyFile (hash, dir, key1, name1, name)
       | NodeFlagged (_, node) ->
            phony_node node name

   (*
    * Intern a string with no escapes.
    *)
   let intern dir name =
      let { base_nodes = nodes } = base in

      (* Try DirPhony first *)
      let name = Lm_filename_util.unescape_string name in
      let key  = Lm_filename_util.normalize_string name in
      let hash = Hash.hash_file dir key in
      let node = NodePhonyDir (hash, dir, key, name) in
         try NodeTable.find nodes node with
            Not_found ->
               (* Try GlobalPhony next *)
               let hash = Hash.hash_global_phony name in
               let node = NodePhonyGlobal (hash, name) in
                  try NodeTable.find nodes node with
                     Not_found ->
                        (* Otherwise, it is a normal file *)
                        let node = new_file dir name in
                           try NodeTable.find nodes node with
                              Not_found ->
                                 base.base_nodes <- NodeTable.add nodes node node;
                                 node

   (*
    * Core node without modifiers.
    *)
   let core node =
      match node with
         NodeFlagged (_, node) ->
            node
       | _ ->
            node

   (*
    * Escape a node.
    *)
   let escape kind node =
      let node =
         match node with
            NodeFlagged (_, node) ->
               node
          | _ ->
               node
      in
         match kind with
            NodeNormal ->
               node
          | NodePhony ->
               raise (Invalid_argument "Omake_node.Node.escape: NodePhony is not allowed")
          | NodeOptional ->
               NodeFlagged (NodeIsOptional, node)
          | NodeExists ->
               NodeFlagged (NodeIsExisting, node)
          | NodeSquashed ->
               NodeFlagged (NodeIsSquashed, node)

   (*
    * Get the name.
    *)
   let phony_name name =
      "<" ^ name ^ ">"

   (*
    * Name of the node.
    *)
   let rec name dir1 node =
      match node with
         NodePhonyGlobal (_, name) ->
            phony_name name
       | NodePhonyDir (_, dir2, _, name) ->
            phony_name (relocate_file dir1 dir2 name)
       | NodePhonyFile (_, dir2, _, name1, name2) ->
            phony_name (relocate_file dir1 dir2 name1 ^ ":" ^ name2)
       | NodeFile (_, dir2, _, name) ->
            relocate_file dir1 dir2 name
       | NodeFlagged (_, node) ->
            name dir1 node

   (*
    * Full name is relative to the cwd.
    *)
   let fullname node =
      name (Dir.cwd ()) node

   let absname node =
      name Dir.root node

   (*
    * Hash code for a node.
    *)
   let rec hash = function
      NodePhonyGlobal (hash, _)
    | NodePhonyDir (hash, _, _, _)
    | NodePhonyFile (hash, _, _, _, _)
    | NodeFile (hash, _, _, _) ->
         hash
    | NodeFlagged (_, node) ->
         hash node

   (*
    * Get the core node.
    *)
   let rec core node =
      match node with
         NodePhonyGlobal _
       | NodePhonyDir _
       | NodePhonyFile _
       | NodeFile _ ->
            node
       | NodeFlagged (_, node) ->
            core node

   (*
    * Kind of the node.
    *)
   let kind = function
      NodePhonyGlobal _
    | NodePhonyDir _
    | NodePhonyFile _ ->
         NodePhony
    | NodeFile _ ->
         NodeNormal
    | NodeFlagged (NodeIsOptional, _) ->
         NodeOptional
    | NodeFlagged (NodeIsExisting, _) ->
         NodeExists
    | NodeFlagged (NodeIsSquashed, _) ->
         NodeSquashed

   (*
    * Phony nodes.
    *)
   let rec is_phony = function
      NodePhonyGlobal _
    | NodePhonyDir _
    | NodePhonyFile _ ->
         true
    | NodeFile _ ->
         false
    | NodeFlagged (_, node) ->
         is_phony node

   let rec is_real = function
      NodePhonyGlobal _
    | NodePhonyDir _
    | NodePhonyFile _
    | NodeFlagged (NodeIsOptional, _)
    | NodeFlagged (NodeIsExisting, _) ->
         false
    | NodeFile _ ->
         true
    | NodeFlagged (NodeIsSquashed, node) ->
         is_real node

   (*
    * Existential flag.
    *)
   let always_exists = function
      NodeFlagged (NodeIsOptional, _)
    | NodeFlagged (NodeIsExisting, _) ->
         true
    | NodeFlagged (NodeIsSquashed, _)
    | NodePhonyGlobal _
    | NodePhonyDir _
    | NodePhonyFile _
    | NodeFile _ ->
         false

   (*
    * Just the tail name.
    *)
   let rec tail node =
      match node with
         NodePhonyGlobal (_, name)
       | NodePhonyDir (_, _, _, name)
       | NodePhonyFile (_, _, _, _, name)
       | NodeFile (_, _, _, name) ->
            name
       | NodeFlagged (_, node) ->
            tail node

   (*
    * Get the name of the directory.
    *)
   let rec dir node =
      match node with
         NodePhonyGlobal _ ->
            null_root
       | NodePhonyDir (_, dir, _, _)
       | NodePhonyFile (_, dir, _, _, _)
       | NodeFile (_, dir, _, _) ->
            dir
       | NodeFlagged (_, node) ->
            dir node

   (*
    * Equality testing.
    *)
   let compare = NodeCompare.compare

   let compare_alpha = NodeCompare.compare_alpha

   let equal node1 node2 =
      compare node1 node2 = 0

   (*
    * Flags.
    *)
   let marshal_flag = function
      NodeIsOptional ->
         Magic NodeIsOptionalMagic
    | NodeIsExisting ->
         Magic NodeIsExistingMagic
    | NodeIsSquashed ->
         Magic NodeIsSquashedMagic

   let unmarshal_flag flag =
      match flag with
         Magic NodeIsOptionalMagic ->
            NodeIsOptional
       | Magic NodeIsExistingMagic ->
            NodeIsExisting
       | Magic NodeIsSquashedMagic ->
            NodeIsSquashed
       | _ ->
            raise MarshalError

   (*
    * Marshaling.
    *)
   let rec marshal node =
      match node with
         NodeFile (i, dir, name1, name2) ->
            List [Magic NodeFileMagic; Int i; Dir.marshal dir; String name1; String name2]
       | NodePhonyGlobal (i, s) ->
            List [Magic NodePhonyGlobalMagic; Int i; String s]
       | NodePhonyDir (i, dir, name1, name2) ->
            List [Magic NodePhonyDirMagic; Int i; Dir.marshal dir; String name1; String name2]
       | NodePhonyFile (i, dir, name1, name2, name3) ->
            List [Magic NodePhonyFileMagic; Int i; Dir.marshal dir; String name1; String name2; String name3]
       | NodeFlagged (flag, node) ->
            List [Magic NodeFlaggedMagic; marshal_flag flag; marshal node]

   let rec unmarshal l =
      match l with
         List [Magic NodeFileMagic; Int i; dir; String name1; String name2] ->
            NodeFile (i, Dir.unmarshal dir, name1, name2)
       | List [Magic NodePhonyGlobalMagic; Int i; String s] ->
            NodePhonyGlobal (i, s)
       | List [Magic NodePhonyDirMagic; Int i; dir; String name1; String name2] ->
            NodePhonyDir (i, Dir.unmarshal dir, name1, name2)
       | List [Magic NodePhonyFileMagic; Int i; dir; String name1; String name2; String name3] ->
            NodePhonyFile (i, Dir.unmarshal dir, name1, name2, name3)
       | List [Magic NodeFlaggedMagic; flag; node] ->
            NodeFlagged (unmarshal_flag flag, unmarshal node)
       | _ ->
            raise MarshalError
end

(*
 * Print the directory, for debugging.
 *)
let pp_print_dir buf dir =
   let root, _, path = path_of_dir dir in
      fprintf buf "%s%s" (**)
         (Lm_filename_util.string_of_root root)
         (flatten_dir path)

(*
 * Print the kind.
 *)
let pp_print_node_kind buf kind =
   let s =
      match kind with
         NodePhony    -> "phony"
       | NodeOptional -> "optional"
       | NodeExists   -> "exists"
       | NodeSquashed -> "squashed"
       | NodeNormal   -> "normal"
   in
      pp_print_string buf s

(*
 * Print the node, for debugging.
 *)
let pp_print_node buf node =
   match Node.kind node with
      NodePhony ->
         fprintf buf "<phony %s>" (Node.fullname node)
    | NodeOptional ->
         fprintf buf "<optional %s>" (Node.fullname node)
    | NodeExists ->
         fprintf buf "<exists %s>" (Node.fullname node)
    | NodeSquashed ->
         fprintf buf "<squash %s>" (Node.fullname node)
    | NodeNormal ->
         pp_print_string buf (Node.fullname node)

let pp_print_string_list buf sources =
   List.iter (fun s -> fprintf buf "@ %s" s) sources

let pp_print_node_set buf set =
   NodeSet.iter (fun node -> fprintf buf "@ %a" pp_print_node node) set

let pp_print_node_table buf table =
   NodeTable.iter (fun node _ -> fprintf buf "@ %a" pp_print_node node) table

let pp_print_node_set_table buf table =
   NodeTable.iter (fun node set ->
         fprintf buf "@ @[<b 3>%a:%a@]" (**)
            pp_print_node node
            pp_print_node_set set) table

let pp_print_node_set_table_opt buf table_opt =
   match table_opt with
      Some table ->
         pp_print_node_set_table buf table
    | None ->
         pp_print_string buf "<none>"

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
